#include <avr/io.h>
#include <avr/interrupt.h>

#include "types.h"
#include "delay.h"
#include "spi.h"
#include "mmc.h"


#define MMC_CS_DDR	DDRB
#define MMC_CS_PIN	PINB
#define MMC_CS_PORT	PORTB
#define PIN_MMC_CS PB2

// look in datasheet for complete list of commands
#define MMC_CMD_0_GO_IDLE				0
#define MMC_CMD_1_SEND_OP_COND			1
#define MMC_CMD_9_SEND_CSD				9
#define MMC_CMD_10_SEND_CID				10
#define MMC_CMD_12_STOP					12
#define MMC_CMD_13_SEND_STATUS			13
#define MMC_CMD_16_BLOCKLEN				16
#define MMC_CMD_17_READ_SINGLE			17
#define MMC_CMD_18_READ_MULTIPLE		18
#define MMC_CMD_24_WRITE_SINGLE			0x18

// bit definitions for R1
#define MMC_R1_IN_IDLE 	0x01
#define MMC_R1B_BUSY_BYTE 0x00

// different defines for tokens etc ...
#define MMC_START_TOKEN_SINGLE  0xfe
#define MMC_START_TOKEN_MULTI   0xfc


// starts the read process of a sector
static void MMC_get_sec_start(u16 sectorh, u16 sectorl);
// gets the next byte from the MMC card
static u08 MMC_get_sec_next(void);
// stop read process of a sector
static void MMC_get_sec_stop(void);

// deselcets the MMC
static void MMC_CS_deselect(void);
//selects the MMC for transmission
static void MMC_CS_select(void);

static void MMC_cleanup(void);


//---------------------------------------


u08 mmc_open=0;

// sends a block of data from the mem over the spi bus
static void spi_io_mem(u08* data, u16 length){
	// transmit 'length' bytes over spi
	while(length){
		spi_io(*data);
		data++;
		length--;
	};
};

// send cmd + arguments + default crc for init command. once in spi mode 
// we dont need crc so we can keep this constant
static void MMC_send_cmd(u08 cmd, u16 datah, u16 datal){
	// default command sequence
	u08 buffer[6] ;
	// fill sequence with our specific data
	buffer[0]=0x40 | cmd;
	buffer[1]=(datah&0xff00)>>8;
	buffer[2]=datah&0xff;
	buffer[3]=(datal&0xff00)>>8;
	buffer[4]=datal&0xff;
	buffer[5]=0x95;
	// dispach data
	spi_io_mem(buffer,6);
};

// gets a 1 byte long R1
static u08 MMC_get_R1(void){
	u08 retval;
	u08 max_errors = 128;
	// wait for first valid response byte
	do{
		retval = spi_io(0xff);
		max_errors--;
	}while(  (retval & 0x80) && (max_errors>0));
	return retval;
};

// waits for the card to send the start data block token
void MMC_wait_for_start_token(u08 max_errors){
  u08 retval;
  do{
    // get a byte from the spi bus
    retval = spi_io(0xff);
    // keep track of the trys
    max_errors--;   
  }while((retval != MMC_START_TOKEN_SINGLE) );
};

/*
#define MMC_CS_select() cbi(MMC_CS_PORT, PIN_MMC_CS)
#define MMC_CS_deselect() sbi(MMC_CS_PORT, PIN_MMC_CS)
*/
// selects the CS for spi xfer
static void MMC_CS_select(void){
	// pull down the MMC CS line
	cbi(MMC_CS_PORT, PIN_MMC_CS);
};

// deselects the CS for spi xfer
static void MMC_CS_deselect(void){
	// pull up the MMC CS card
	sbi(MMC_CS_PORT, PIN_MMC_CS);
};


// stops the MMC transmission and sends the clock cycles needed by the mmc for cleanup
static void MMC_cleanup(void){
	// deselect the MMC card
	MMC_CS_deselect();
	// pulse the SCK 16 times
	spi_io(0xff);
	spi_io(0xff);
};



// starts the read process of a sector
static void MMC_get_sec_start(u16 sectorh, u16 sectorl){
	spi_io(0xff);
	spi_io(0xff);
	// turn sectors into byte addr
	sectorh = (sectorh << 9) + (sectorl >> 7);
	sectorl = sectorl << 9;
	// select card
	MMC_CS_select();
	mmc_open = 1;
	// tell the MMC card that we want to know its status
	MMC_send_cmd(MMC_CMD_17_READ_SINGLE, sectorh, sectorl);
	// get the response
	if (MMC_get_R1()) {
		MMC_cleanup();
		return;
	}
	// wait till the mmc starts sending data
	MMC_wait_for_start_token(255);
};

// gets the next byte from the MMC card
static u08 MMC_get_sec_next(void){
	u08 tmp = spi_io(0xff);
	return tmp;
};

// stop read process of a sector
static void MMC_get_sec_stop(void){
	if (mmc_open) {
		// get the 2 CRC bytes
		spi_io(0xff);
		spi_io(0xff);
		// give enough time
		MMC_cleanup();
		mmc_open = 0;
	}
};




//-------------------------------------------------------------
// High Level MMC-Funktionen

u16 sect;
u16 bufpos=0;


// Lesen beginnen
void mmc_load_start(u16 s) {
	sect = s;
	MMC_get_sec_start(0, (u16)(s));
	bufpos = 0;
}


// Byte lesen
u08 mmc_fetch_byte(void) {
	if (bufpos == 512) {
		MMC_get_sec_stop();
		mmc_load_start(++sect);
	}
	bufpos++;
	return MMC_get_sec_next();
}


// Lesen, um die 512 voll zu machen
void mmc_complete_read(void) {
	while (bufpos < 512) {
		MMC_get_sec_next();
		bufpos++;
	}
	MMC_get_sec_stop();
}


void mmc_read_block(u08 *buf, u16 cnt) {
	while (cnt--) {
		*buf = mmc_fetch_byte();
		buf++;
	}
}


